/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.web;

import de.otto.mongodb.profiler.AverageMeasure;
import de.otto.mongodb.profiler.collection.CollectionProfile;
import de.otto.mongodb.profiler.web.util.DataSizeFormat;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.util.Date;
import java.util.Locale;
import java.util.NoSuchElementException;

public abstract class CollectionProfileViewModel {

    public static final MathContext ABSOLUTE_AVERAGE_MATH_CONTEXT = new MathContext(16, RoundingMode.HALF_UP);

    public static BigDecimal averageAbsolute(AverageMeasure measure) {
        return measure.getAverage(ABSOLUTE_AVERAGE_MATH_CONTEXT);
    }

    public static final MathContext PADDING_FACTOR_AVERAGE_MATH_CONTEXT = MathContext.DECIMAL64;

    public static BigDecimal averagePaddingFactor(AverageMeasure measure) {
        return measure.getAverage(PADDING_FACTOR_AVERAGE_MATH_CONTEXT);
    }

    private final Locale locale;
    private final DateFormat dateFormat;
    private final NumberFormat integerNumberFormat;

    protected CollectionProfileViewModel(final Locale locale) {
        this.locale = locale;
        this.dateFormat = DateFormat.getDateTimeInstance(DateFormat.DEFAULT, DateFormat.DEFAULT, locale);
        this.integerNumberFormat = NumberFormat.getIntegerInstance(locale);
        this.integerNumberFormat.setGroupingUsed(true);
        this.integerNumberFormat.setMaximumFractionDigits(0);
        this.integerNumberFormat.setMinimumFractionDigits(0);
    }

    protected Locale getLocale() {
        return locale;
    }

    public static CollectionProfile.Mark getLastMark(CollectionProfile profile) {
        try {
            return profile.getMarks().getLast();
        } catch (NoSuchElementException e) {
            return null;
        }
    }

    public String getLastDocunentCount(CollectionProfile.Mark mark) {
        return mark != null ? integerNumberFormat.format(mark.getDocumentCount()) : null;
    }

    public static String getLastPaddingFactor(CollectionProfile.Mark mark) {
        return mark != null ? String.valueOf(mark.getPaddingFactor()) : null;
    }

    public String getLastDataSize(CollectionProfile.Mark mark) {
        return mark != null ? DataSizeFormat.format(mark.getDataSize(), locale) : null;
    }

    public String getLastDataStorageSize(CollectionProfile.Mark mark) {
        return mark != null ? DataSizeFormat.format(mark.getDataStorageSize(), locale) : null;
    }

    public String getLastSampleTime(CollectionProfile.Mark mark) {
        return mark != null ? dateFormat.format(new Date(mark.getTime())) : null;
    }

    public String formatBytes(long bytes) {
        return DataSizeFormat.format(bytes, locale);
    }

    public String formatInteger(int value) {
        return integerNumberFormat.format(value);
    }
}
