/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.web;

import de.otto.mongodb.profiler.op.*;
import org.joda.time.DateTime;

import java.math.RoundingMode;
import java.text.Collator;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Locale;

import static java.util.Collections.sort;

public class OpProfilesPanelViewModel extends OpProfileViewModel {

    public static final Collator COLLATOR = Collator.getInstance(Locale.ENGLISH);

    public static final Comparator<OpProfile> QUERY_PROFILE_COMPARATOR = new Comparator<OpProfile>() {
        @Override
        public int compare(OpProfile left, OpProfile right) {

            if (left == right) {
                return 0;
            }

            if (left == null) {
                return -1;
            }

            if (right == null) {
                return 1;
            }

            int diff = COLLATOR.compare(left.getNs(), right.getNs());

            if (diff != 0) {
                return diff;
            }

            diff = COLLATOR.compare(left.getType(), right.getType());

            if (diff != 0) {
                return diff;
            }

            return COLLATOR.compare(left.getId(), right.getId());
        }
    };

    public static enum RenderType {
        PANEL,
        TABLE;
    }

    private final String connectionId;
    private final String databaseName;
    private final List<? extends OpProfile> queryProfiles;
    private final boolean profilingEnabled;
    private final DateTime profilingSince;
    private final OpProfiler.ProfilingLevel profilingLevel;
    private final RenderType renderType;
    private final long totalEventCount;
    private final long totalExplanationCount;
    private final NumberFormat percentFormat;
    private final boolean outlierConfigurationEnabled;

    public OpProfilesPanelViewModel(final String connectionId,
                                    final String databaseName,
                                    final OpProfiler profiler,
                                    final RenderType renderType,
                                    final Locale locale) {
        super(locale);
        this.connectionId = connectionId;
        this.databaseName = databaseName;
        this.queryProfiles = new ArrayList<>(profiler.getProfiles());
        sort(this.queryProfiles, QUERY_PROFILE_COMPARATOR);
        this.profilingEnabled = profiler.isProfiling();
        this.profilingSince = profiler.getProfilingStarted();
        this.profilingLevel = profiler.getProfilingLevel();
        this.renderType = renderType;
        this.outlierConfigurationEnabled = profiler.getOutlierConfiguration() != null;

        this.percentFormat = NumberFormat.getPercentInstance(Locale.ENGLISH);
        this.percentFormat.setMaximumFractionDigits(0);
        this.percentFormat.setMinimumFractionDigits(0);
        this.percentFormat.setRoundingMode(RoundingMode.UP);

        // Totals
        long totalEventCount = 0;
        long totalExplanationCount = 0;
        for (OpProfile profile : queryProfiles) {
            totalEventCount += profile.getEventCount();
            if (profile instanceof QueryProfile) {
                totalExplanationCount += ((QueryProfile) profile).getExplanationCount();
            }
        }
        this.totalEventCount = totalEventCount;
        this.totalExplanationCount = totalExplanationCount;
    }

    public long getTotalEventCount() {
        return totalEventCount;
    }

    public long getTotalExplanationCount() {
        return totalExplanationCount;
    }

    public String getEventCountProportion(OpProfile profile) {
        final double percent = (((double) profile.getEventCount()) / ((double) totalEventCount));
        return percentFormat.format(percent);
    }

    public String getRenderType() {
        return renderType.name();
    }

    public String getConnectionId() {
        return connectionId;
    }

    public String getDatabaseName() {
        return databaseName;
    }

    public Iterable<? extends OpProfile> getProfiles() {
        return queryProfiles;
    }

    public boolean isProfilingEnabled() {
        return profilingEnabled;
    }

    public DateTime getProfilingSince() {
        return profilingSince;
    }

    public OpProfiler.ProfilingLevel getProfilingLevel() {
        return profilingLevel;
    }

    public static Object getContext(OpProfile profile) {
        return OpProfileType.contextOf(profile);
    }

    public static boolean hasSevereIssue(OpProfile profile) {
        return hasIssue(profile, ExplanationIssue.Severity.SEVERE);
    }

    public static boolean hasWarningIssue(OpProfile profile) {
        return hasIssue(profile, ExplanationIssue.Severity.WARNING)
                || hasIssue(profile, ProfileIssue.Severity.WARNING);
    }

    public static boolean hasIssue(OpProfile profile, ExplanationIssue.Severity severity) {

        if (profile instanceof QueryProfile == false) {
            return false;
        }

        final Explanation lastExplanation = ((QueryProfile) profile).getLastExplanation();
        if (lastExplanation != null) {
            for (ExplanationIssue issue : lastExplanation.getIssues()) {
                if (severity.equals(issue.getSeverity())) {
                    return true;
                }
            }
        }
        return false;
    }

    public static boolean hasIssue(OpProfile profile, ProfileIssue.Severity severity) {

        for (ProfileIssue issue : profile.getIssues()) {
            if (severity.equals(issue.getSeverity())) {
                return true;
            }
        }
        return false;
    }

    public boolean isOutlierConfigurationEnabled() {
        return outlierConfigurationEnabled;
    }
}
