/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.web.util;

import org.bson.BSONObject;
import org.bson.types.*;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

public class BSONFormat {

    public static BSONFormat getHtmlInstance() {
        return new BSONFormat(true, 0);
    }

    public static BSONFormat getTextInstance(int intendation) {
        return new BSONFormat(false, intendation);
    }

    private static String spaces(int spaces) {
        String s = "";
        while (spaces-- > 0) {
            s += ' ';
        }
        return s;
    }

    private final boolean html;
    private final int intendation;
    private final DateFormat dateFormat;

    private BSONFormat(boolean html, int intendation) {
        this.html = html;
        this.intendation = intendation;
        this.dateFormat = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z", Locale.ENGLISH);
    }

    public String format(BSONObject bson) {
        final StringBuilder sb = new StringBuilder();
        format(bson, sb, 1);
        return sb.toString();
    }

    private void format(BSONObject bson, StringBuilder sb, int level) {

        final Iterator<String> keys = bson.keySet().iterator();

        if (!keys.hasNext()) {
            sb.append("{}");
            return;
        }

        if (bson instanceof List) {
            sb.append('[');
        } else {
            sb.append('{');
        }

        if (html) {
            sb.append("<ul>");
        } else {
            sb.append('\n');
        }

        final String spaces = spaces(level * intendation);
        while (keys.hasNext()) {

            final String key = keys.next();

            if (html) {
                sb.append("<li>");
            } else {
                sb.append(spaces);
            }

            if (bson instanceof List == false) {
                sb.append(key).append(": ");
            }

            final String comment = formatValue(bson.get(key), sb, level);

            if (keys.hasNext()) {
                sb.append(',');
            }

            if (comment != null) {
                sb.append(" // ").append(comment);
            }

            if (html) {
                sb.append("</li>");
            } else {
                sb.append('\n');
            }
        }

        if (html) {
            sb.append("</ul>");
        }

        if (!html && level > 1) {
            sb.append(spaces((level - 1) * intendation));
        }

        if (bson instanceof List) {
            sb.append(']');
        } else {
            sb.append('}');
        }
    }


    private String formatValue(Object value, StringBuilder sb, int level) {

        if (value == null) {
            sb.append("null");
        } else if (value instanceof String) {
            sb.append('\'').append(value).append('\'');
        } else if (value instanceof Number) {
            sb.append(value);
        } else if (value instanceof Boolean) {
            sb.append(value);
        } else if (value instanceof Date) {
            sb.append("new Date('").append(dateFormat.format(((Date) value))).append("')");
        } else if (value instanceof BSONObject) {
            format(((BSONObject) value), sb, level + 1);
        } else if (value instanceof Pattern) {
            final Pattern p = ((Pattern) value);
            sb.append("new Pattern('").append(p.pattern()).append("')");
        } else if (value instanceof ObjectId) {
            sb.append("new ObjectId('").append(((ObjectId) value).toStringMongod()).append("')");
        } else if (value instanceof Symbol) {
            sb.append("new Symbol('").append(((Symbol) value).getSymbol()).append("')");
        } else if (value instanceof Code) {
            sb.append("new Code('").append(((Code) value).getCode()).append("')");
        } else if (value instanceof CodeWScope) {
            sb.append("new CodeWScope('").append(((CodeWScope) value).getCode()).append("')");
        } else if (value instanceof Binary) {
            final Binary b = ((Binary) value);
            sb.append("new Binary(...)");
            return String.format("%s, type %s", DataSizeFormat.format(b.length(), Locale.ENGLISH), b.getType());
        } else if (value instanceof byte[]) {
            sb.append("new Binary(...)");
            return String.format("%s, type %s", DataSizeFormat.format(((byte[]) value).length, Locale.ENGLISH), 0);
        } else if (value instanceof BSONTimestamp) {
            final BSONTimestamp ts = ((BSONTimestamp) value);
            sb.append("new BSONTimestamp('").append(ts.getTime()).append(", ").append(ts.getInc()).append("')");
        }

        return null;
    }
}
