/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler;

import org.joda.time.DateTime;
import org.joda.time.Interval;

import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * A profiler that operates asynchronously, using it's own scheduling and threading. Such a profiler must be stoppable
 * and continuable.
 *
 * @author Robert Gacki
 */
public interface AsyncProfiler extends Profiler {

    /**
     * Starts or continues the profiling. Returns {@code true} if the profiling is being continued. Returns
     * {@code false} if continuation is not possible or if the profiling is already continuing. Throws an
     * {@link IllegalStateException} if the profiler is shut down.
     *
     * @return {@code true} if the profiling is continuing due to this call, else {@code false}
     * @throws IllegalStateException if the profiler is shut down
     */
    boolean continueProfiling() throws IllegalStateException;

    /**
     * Starts or continues the profiling and sets a duration. After that duration, the profiling is stopped
     * automatically. Returns the DateTime for when the profiling is stopped. Returns {@code null} if the
     * profiling cannot be continued or is already continuing. Throws an {@link IllegalStateException} if
     * the profiler is shut down.
     *
     * @param duration the duration until the profiling stops
     * @param timeUnit the time unit of the duration
     * @return the point in time when the profiling will be stopped or {@code null} if this call had no affect
     * @throws IllegalStateException if the profiler is shut down
     */
    DateTime continueProfiling(long duration, TimeUnit timeUnit) throws IllegalStateException;

    /**
     * Returns the current timeout for stopping the profiling, if set. Returns {@code null} if there is no timeout set.
     *
     * @return the timeout for stopping the profiling if set
     */
    DateTime getTimeout();

    /**
     * Stops the profiling. Returns {@code true} if the profiling is stopped upon this call. Returns {@code false} if
     * the profiling is already stopped or if the call had no affect.
     *
     * @return {@code true} if the profiling is stopped upon this call
     */
    boolean stopProfiling();

    /**
     * Returns {@code true} if the profiling is active. Else, {@code false} is returned,
     *
     * @return {@code true} if the profiling is active, else {@code false}
     */
    boolean isProfiling();

    /**
     * Returns the date and time when the profiling has started. Returns {@code null} if the Profiler is
     * not profiling.
     *
     * @return the DateTime or {@code null}
     */
    DateTime getProfilingStarted();

    /**
     * Returns a list of intervals the profiler has run in the past. Returns an empty list if the profiler
     * did not run or was not stopped once.
     *
     * @return the list of intervals
     */
    List<Interval> getProfilingRuntimes();

}
