/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Collection;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * Creates a statistical summary of a stream of values. This type is immutable. Adding a value will return a new updated
 * instance of the measurement. It also takes an OutlierStrategy in order to detect and handle outlying values.
 *
 * @author Robert Gacki
 */
public class AverageMeasure {

    /**
     * The default value, which is a {@link BigDecimal#ZERO}.
     */
    public static final BigDecimal DEFAULT_VALUE = BigDecimal.ZERO;

    /**
     * The default value for the minimum: {@value #DEFAULT_MIN}
     */
    public static final double DEFAULT_MIN = 0.0;

    /**
     * The default value for the maximum: {@value #DEFAULT_MAX}
     */
    public static final double DEFAULT_MAX = 0.0;

    /**
     * The default value count: {@value #DEFAULT_COUNT}
     */
    public static final long DEFAULT_COUNT = 0L;

    /**
     * The strategy to choose whether a value is an outlier.
     *
     * @author Robert Gacki
     */
    public interface OutlierStrategy {

        /**
         * Determines, whether a value is an outlier. Returns {@code true} if the value is to be treated as an outlier.
         * Returns {@code false} if the value is to be treated as normal.
         *
         * @param value   the value to check
         * @param current the current measure
         * @return {@code true} if the value is an outlier, else {@code false}
         */
        boolean isOutlier(double value, AverageMeasure current);

    }

    /**
     * A type to track outliers. The type consists of the value and the measure at the time, the value was determined to
     * be an outlier.
     *
     * @author Robert Gacki
     */
    public static class Outlier {

        private final double value;
        private final AverageMeasure measure;
        private final long time;

        public Outlier(final double value, final AverageMeasure measure, final long time) {
            this.value = value;
            this.measure = measure;
            this.time = time;
        }

        public double getValue() {
            return value;
        }

        public AverageMeasure getMeasure() {
            return measure;
        }

        public long getTime() {
            return time;
        }
    }

    /**
     * A strategy that captures the outliers, determined by another strategy that is wrapped and adds them to a
     * collection.
     *
     * @author Robert Gacki
     */
    public static class CapturingOutlierStrategy implements OutlierStrategy {

        private final OutlierStrategy strategy;
        private final Collection<Outlier> outliers;

        /**
         * Creates an instance that takes a strategy and an external collection. The collection is used by reference.
         *
         * @param strategy the strategy
         * @param outliers the collection to store the outliers in
         */
        public CapturingOutlierStrategy(final OutlierStrategy strategy,
                                        final Collection<Outlier> outliers) {
            this.strategy = checkNotNull(strategy);
            this.outliers = checkNotNull(outliers);
        }

        /**
         * In addition to the behavior of the wrapped strategy, this method adds outliers to the collection.
         *
         * @see OutlierStrategy#isOutlier(double, AverageMeasure)
         */
        @Override
        public boolean isOutlier(double value, AverageMeasure current) {
            if (strategy.isOutlier(value, current)) {
                this.outliers.add(new Outlier(value, current, System.currentTimeMillis()));
                return true;
            }
            return false;
        }
    }

    /**
     * A strategy which requires outliers to be larger than two times of the current mean value.
     *
     * @author Robert Gacki
     */
    public static final OutlierStrategy TWO_TIMES_MEAN_VALUE = new OutlierStrategy() {

        private final MathContext mathContext = new MathContext(16, RoundingMode.HALF_UP);
        private final BigDecimal two = new BigDecimal("2.0");

        @Override
        public boolean isOutlier(double value, AverageMeasure current) {
            return value > current.getAverage(mathContext).multiply(two).doubleValue();
        }
    };

    private final BigDecimal value;
    private final double min;
    private final double max;
    private final long count;

    private final boolean refresh;

    private final OutlierStrategy outlierStrategy;
    private final boolean ignoreOutliers;
    private final long outlierCount;

    /**
     * Creates a new average measure that does include outliers.
     */
    public AverageMeasure() {
        this(true);
    }

    /**
     * Creates a new average measure with a specific strategy for outlying values.
     *
     * @param outlierStrategy the strategy to use to determine outliers
     * @param ignoreOutliers  {@code true} if outliers are to be ignored for the average calculation
     */
    public AverageMeasure(final OutlierStrategy outlierStrategy,
                          final boolean ignoreOutliers) {
        this(true, outlierStrategy, ignoreOutliers);
    }

    /**
     * Creates a new average measure that does specify whether to refresh the values at the call of {@link #add(double)}.
     * If not refreshed, the average is built upon the defaults {@link #DEFAULT_VALUE}, {@link #DEFAULT_COUNT},
     * {@link #DEFAULT_MIN} and {@link #DEFAULT_MAX}.
     *
     * @param refresh {@code true} if
     */
    public AverageMeasure(final boolean refresh) {
        this(DEFAULT_VALUE, DEFAULT_MIN, DEFAULT_MAX, DEFAULT_COUNT, refresh, null, false, 0L);
    }

    public AverageMeasure(final boolean refresh,
                          final OutlierStrategy outlierStrategy,
                          final boolean ignoreOutliers) {
        this(DEFAULT_VALUE, DEFAULT_MIN, DEFAULT_MAX, DEFAULT_COUNT, refresh, outlierStrategy, ignoreOutliers, 0L);
    }

    public AverageMeasure(final BigDecimal value,
                          final double min,
                          final double max,
                          final long count) {
        this(value, min, max, count, false, null, false, 0L);
    }

    public AverageMeasure(final BigDecimal value,
                          final double min,
                          final double max,
                          final long count,
                          final OutlierStrategy outlierStrategy,
                          final boolean ignoreOutliers) {
        this(value, min, max, count, false, outlierStrategy, ignoreOutliers, 0L);
    }

    private AverageMeasure(final BigDecimal value,
                           final double min,
                           final double max,
                           final long count,
                           final boolean refresh,
                           final OutlierStrategy outlierStrategy,
                           final boolean ignoreOutliers,
                           final long outlierCount) {
        this.value = value;
        this.min = min;
        this.max = max;
        this.count = count;
        this.refresh = refresh;
        this.outlierStrategy = outlierStrategy;
        this.ignoreOutliers = ignoreOutliers;
        this.outlierCount = outlierCount;
    }

    public AverageMeasure add(double amount) {

        if (count == Long.MAX_VALUE) {
            throw new RuntimeException("Unable to add further values. Value count reached Long.MAX_VALUE!");
        }

        amount = Math.abs(amount);

        if (refresh) {
            return new AverageMeasure(
                    new BigDecimal(String.valueOf(amount)),
                    amount,
                    amount,
                    1L,
                    outlierStrategy,
                    ignoreOutliers);
        }

        final boolean outlier = outlierStrategy != null && outlierStrategy.isOutlier(amount, this);

        if (outlier && ignoreOutliers) {
            return new AverageMeasure(
                    value,
                    min,
                    max,
                    1L,
                    false,
                    outlierStrategy,
                    ignoreOutliers,
                    outlierCount + 1L);
        }

        return new AverageMeasure(
                value.add(new BigDecimal(String.valueOf(amount))),
                amount < min ? amount : min,
                amount > max ? amount : max,
                count + 1L,
                false,
                outlierStrategy,
                ignoreOutliers,
                outlier ? outlierCount + 1 : outlierCount);
    }

    /**
     * Returns the average value.
     *
     * @return the average value
     */
    public BigDecimal getAverage() {
        return count > 1L ? value.divide(BigDecimal.valueOf(count)) : value;
    }

    /**
     * Returns the average value for a certain math context.
     *
     * @param mathContext the math context to use
     * @return the average value within the math context
     */
    public BigDecimal getAverage(MathContext mathContext) {
        return count > 1L ? value.divide(BigDecimal.valueOf(count), mathContext) : value;
    }

    /**
     * Returns the minimum value.
     *
     * @return the minimum value
     */
    public double getMin() {
        return min;
    }

    /**
     * Returns the maximum value.
     *
     * @return the maximum value
     */
    public double getMax() {
        return max;
    }

    /**
     * Returns the amount of events.
     *
     * @return the amount of events
     */
    public long getCount() {
        return count;
    }

    /**
     * Returns the amount of outliers determined by the outlier strategy that is used.
     *
     * @return the amount of outliers
     */
    public long getOutlierCount() {
        return outlierCount;
    }

    /**
     * Returns the {@link OutlierStrategy} that is used to determine outlying values.
     *
     * @return the strategy
     */
    public OutlierStrategy getOutlierStrategy() {
        return outlierStrategy;
    }

    /**
     * Determines whether outlying values are ignored by the measurement. The default is {@code false}, meaning outlying
     * values are not ignored.
     *
     * @return {@code true} if outlying values are ignored
     */
    public boolean isIgnoreOutliers() {
        return ignoreOutliers;
    }

    @Override
    public String toString() {
        return "AverageMeasure{" +
                "value=" + value +
                ", min=" + min +
                ", max=" + max +
                ", count=" + count +
                ", refresh=" + refresh +
                ", outlierStrategy=" + outlierStrategy +
                ", ignoreOutliers=" + ignoreOutliers +
                ", outlierCount=" + outlierCount +
                "} " + super.toString();
    }
}
