/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler;


import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * <p>A sampler that helps to reduce the amount of samples (observations) by an equal time windows. An instance is not
 * thread-safe and shall not be re-used.</p>
 * <p/>
 * <p>After instantiation, the stream of chronologically subsequent samples is added via {@link #add(long, Object)}.
 * The sampling is finished by calling {@link #finish()} which returns the reduced samples.</p>
 *
 * @param <E> the origin sample type
 * @param <S> the type of the result sample
 * @author Robert Gacki
 */
public class ChronoSampler<E, S> {

    /**
     * A function that implements the method of sample reduction.
     *
     * @param <E> the origin sample type
     * @param <S> the type of the result sample
     * @author Robert Gacki
     */
    public interface Reduction<E, S> {

        /**
         * Returns a new sample for a certain time window that is calculated by a chunk of samples that fit into it.
         * May return {@code null} if a sample could not be created.
         *
         * @param time  the upper bound of the time window
         * @param marks the data that fits into this window
         * @return the new sample or {@code null}
         */
        S reduce(long time, Deque<E> marks);

    }

    private final long window;
    private final TimeUnit windowUnit;
    private final Reduction<E, S> reduction;
    private final Long from;
    private final List<S> samples;
    private Deque<E> chunk;
    private Long currentWindow;
    private Long nextWindow;
    private Long last;

    /**
     * Creates a new sampler.
     *
     * @param window     the window size
     * @param windowUnit the time unit of the window size
     * @param reduction  the reduction function
     */
    public ChronoSampler(final long window,
                         final TimeUnit windowUnit,
                         final Reduction<E, S> reduction) {
        this(window, windowUnit, reduction, null);
    }

    /**
     * Creates a new sampler.
     *
     * @param window     the window size
     * @param windowUnit the time unit of the window size
     * @param reduction  the reduction function
     * @param from       the minimum time of samples to accept
     */
    public ChronoSampler(final long window,
                         final TimeUnit windowUnit,
                         final Reduction<E, S> reduction,
                         final Long from) {
        this.window = window;
        this.windowUnit = checkNotNull(windowUnit);
        this.currentWindow = null;
        this.nextWindow = null;
        this.reduction = checkNotNull(reduction);
        this.from = from;
        this.samples = new ArrayList<>();
        this.chunk = new LinkedList<>();
        this.last = null;
    }

    private boolean generateSample(final long now) {
        if (nextWindow == null) {
            nextWindow = Long.valueOf(now);
        }
        if (now >= nextWindow.longValue()) {
            currentWindow = nextWindow;
            final long millisToNextWindow = windowUnit.toMillis(window);
            nextWindow = Long.valueOf(now + millisToNextWindow);
            return true;
        }
        return false;
    }

    /**
     * Adds a sample of a certain time to the sampler. The time must not be lower than that of the previous invocation.
     *
     * @param time   the time of the sample
     * @param sample the sample to add to the reduction
     * @return the sampler instance
     */
    public ChronoSampler<E, S> add(long time, E sample) {

        checkNotNull(sample);

        if (from != null && time < from.longValue()) {
            return this;
        }

        if (last != null && time < last.longValue()) {
            throw new IllegalArgumentException("Marks must increase chronologically!");
        }

        // First sample
        final boolean firstSample = last == null;
        if (firstSample) {
            chunk.add(sample);
        }

        last = Long.valueOf(time);

        if (generateSample(time)) {
            final S reducedSample = reduction.reduce(currentWindow, chunk);
            if (reducedSample != null) {
                samples.add(reducedSample);
            }
            chunk = new LinkedList<>();
        }

        if (!firstSample) {
            chunk.add(sample);
        }

        return this;
    }

    /**
     * Finishes the sampling and returns the list of reduced samples.
     *
     * @return the list of reduced samples
     */
    public List<S> finish() {

        if (chunk.isEmpty()) {
            return samples;
        }

        final long time = last;
        final S sample = reduction.reduce(time, chunk);
        samples.add(sample);
        chunk = new LinkedList<>();

        return samples;
    }
}
