/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler;

import com.mongodb.MongoClientOptions;
import com.mongodb.ReadPreference;
import com.mongodb.ServerAddress;

import java.net.UnknownHostException;
import java.util.LinkedHashSet;
import java.util.Set;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * The configuration for creating a MongoDB server connection. The {@link ProfiledConnectionConfiguration.Builder}
 * should be used to build the configuration.
 *
 * @author Robert Gacki
 */
public class ProfiledConnectionConfiguration {

    /**
     * The builder that creates a ProfiledConnectionConfiguration.
     *
     * @author Robert Gacki
     */
    public static class Builder {

        private final Set<ServerAddress> serverAddresses;
        private final MongoClientOptions.Builder clientOptions;

        protected Builder() {
            serverAddresses = new LinkedHashSet<>();
            clientOptions = MongoClientOptions
                    .builder()
                    .alwaysUseMBeans(false)
                    .autoConnectRetry(true)
                    .connectionsPerHost(50)
                    .connectTimeout(5_000)
                    .maxWaitTime(0)
                    .readPreference(ReadPreference.primary())
                    .socketKeepAlive(true)
                    .socketTimeout(0);
        }

        /**
         * Adds a host and the port to a set. Tries to resolve the host. Throws an UnknownHostException if the host
         * could not be resolved.
         *
         * @param hostname the hostname or IP address
         * @param port     the port
         * @return the builder instance
         * @throws UnknownHostException if the host could not be resolved
         */
        public Builder addHost(String hostname, int port) throws UnknownHostException {
            serverAddresses.add(new ServerAddress(hostname, port));
            return this;
        }

        /**
         * Adds a host a set. Uses the MongoDB standard port. Tries to resolve the host. Throws an UnknownHostException
         * if the host could not be resolved.
         *
         * @param hostname the hostname or IP address
         * @return the builder instance
         * @throws UnknownHostException if the host could not be resolved
         */
        public Builder addHost(String hostname) throws UnknownHostException {
            serverAddresses.add(new ServerAddress(hostname));
            return this;
        }

        /**
         * Creates the configuration.
         *
         * @return the configuration
         */
        public ProfiledConnectionConfiguration toConfiguration() {
            return new ProfiledConnectionConfiguration(serverAddresses, clientOptions.build());
        }
    }

    /**
     * Returns a new builder instance to create a configuration.
     *
     * @return the builder instance
     */
    public static Builder build() {
        return new Builder();
    }

    private final Set<ServerAddress> serverAddresses;
    private final MongoClientOptions clientOptions;

    protected ProfiledConnectionConfiguration(final Set<ServerAddress> serverAddresses,
                                              final MongoClientOptions clientOptions) {

        this.serverAddresses = checkNotNull(serverAddresses);
        this.clientOptions = checkNotNull(clientOptions);
    }

    /**
     * Returns a set of server addresses to connect to. The addresses must be part of the same MongoDB cluster / replica
     * set.
     *
     * @return the set of server addresses
     */
    public Set<ServerAddress> getServerAddresses() {
        return serverAddresses;
    }

    /**
     * Returns the client options to use when connecting to the MongoDB.
     *
     * @return the client options
     */
    public MongoClientOptions getClientOptions() {
        return clientOptions;
    }
}
