/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler;

import de.otto.mongodb.profiler.collection.CollectionProfiler;
import de.otto.mongodb.profiler.op.OpProfiler;

import java.util.List;

/**
 * A service that manages the lifecycle of connections, databases and profilers.
 *
 * @author Robert Gacki
 */
public interface ProfilerService {

    /**
     * Creates a new connection.
     *
     * @param configuration the configuration used to create the connection
     * @return the reference to the created connection
     */
    ProfiledConnection createConnection(String name, ProfiledConnectionConfiguration configuration)
            throws InvalidConnectionNameException;

    /**
     * Returns a list of connections. Returns an empty list when there is no connection available.
     *
     * @return a list of connections which may be empty
     */
    List<? extends ProfiledConnection> getConnections();

    /**
     * Returns a connection specified by its ID. Returns {@code null} if the connection does not exist.
     *
     * @param id the ID of the connection to return
     * @return the connection or {@code null} if not found
     */
    ProfiledConnection getConnection(String id);

    /**
     * Destroys a connection.
     *
     * @param connection the reference to connection to destroy
     */
    void destroyConnection(ProfiledConnection connection);

    /**
     * Adds a database to be profiled and associates it with the connection. If the database does not exist on the
     * server associated with the connection, a {@link DatabaseDoesNotExistException} is thrown. If the database
     * requires authentication an {@link AuthenticationException} might be thrown, because this method does not do an
     * authentication. If there is already a database configured, a {@link DatabaseAlreadyConfiguredException} is thrown.
     *
     * @param connection the connection the database is associated with
     * @param name       the name of the database
     * @return a reference to the database that is being profiled
     * @throws DatabaseDoesNotExistException if the database does not exist on the server
     * @throws AuthenticationException       if authentication is required
     * @throws ConnectivityException         if a connectivity problem occurred
     */
    ProfiledDatabase addDatabase(ProfiledConnection connection, String name)
            throws DatabaseDoesNotExistException, DatabaseAlreadyConfiguredException, AuthenticationException, ConnectivityException;

    /**
     * Adds a database to be profiled and associates it with the connection. If the database does not exist on
     * the server associated with the connection, a {@link DatabaseDoesNotExistException} is thrown. If both,
     * username and password are not {@code null}, an authentication is executed. An {@link AuthenticationException}
     * may be thrown if either the database does not support authentication or if the authentication failed.  If there
     * is already a database configured, a {@link DatabaseAlreadyConfiguredException} is thrown.
     *
     * @param connection the connection the database is associated with
     * @param name       the name of the database
     * @param username   (optional) the username to authenticate with
     * @param password   (optional) the password to authenticate with
     * @return a reference to the database that is being profiled
     * @throws DatabaseDoesNotExistException if the database does not exist on the server
     * @throws AuthenticationException       if authentication failed or is not supported
     * @throws ConnectivityException         if a connectivity problem occurred
     */
    ProfiledDatabase addDatabase(ProfiledConnection connection, String name, String username, char[] password)
            throws DatabaseDoesNotExistException, DatabaseAlreadyConfiguredException, AuthenticationException, ConnectivityException;

    /**
     * Returns a list of database references that are associated with the connection reference. Returns
     * an empty list, if no databases are associated.
     *
     * @param connection the connection to return the databases for
     * @return the databases references or an empty list
     */
    List<? extends ProfiledDatabase> getDatabases(ProfiledConnection connection);

    /**
     * Returns a database that is associated with a connection for its name. Returns {@code null} if there is no
     * database associated.
     *
     * @param connection the connection to return the associated database for
     * @param name       the name of the database
     * @return the database reference that is associated with the connection or {@code null}
     */
    ProfiledDatabase getDatabase(ProfiledConnection connection, String name);

    /**
     * Removes a database.
     *
     * @param database the database to remove
     */
    void removeDatabase(ProfiledDatabase database);

    /**
     * Returns the {@link CollectionProfiler} for a database reference.
     *
     * @param database the database reference
     * @return the CollectionProfiler
     */
    CollectionProfiler getCollectionProfiler(ProfiledDatabase database);

    /**
     * Returns the {@link OpProfiler} for a database reference.
     *
     * @param database the database reference
     * @return the OpProfiler
     */
    OpProfiler getOpProfiler(ProfiledDatabase database);

}
