/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.collection;

import de.otto.mongodb.profiler.AverageMeasure;

import java.util.Deque;

/**
 * The profile of a collection.
 *
 * @author Robert Gacki
 */
public interface CollectionProfile {

    /**
     * A mark (sample) that contains collection data of a specific point in time.
     *
     * @author Robert Gacki
     */
    public static class Mark {

        private final long time;
        private final double paddingFactor;
        private final long dataSize;
        private final long dataStorageSize;
        private final long documentCount;

        /**
         * Creates a new mark (sample).
         *
         * @param time            the UNIX timestamp the sample was created
         * @param paddingFactor   the padding factor of the collection
         * @param dataSize        the size of the data in bytes
         * @param dataStorageSize the size of the storage (reserved space, extents) in bytes
         * @param documentCount   the document count
         */
        public Mark(final long time,
                    final double paddingFactor,
                    final long dataSize,
                    final long dataStorageSize,
                    final long documentCount) {
            this.time = time;
            this.paddingFactor = paddingFactor;
            this.dataSize = dataSize;
            this.dataStorageSize = dataStorageSize;
            this.documentCount = documentCount;
        }

        /**
         * Returns the UNIX timestamp the sample was created at.
         *
         * @return the UNIX timestamp
         */
        public long getTime() {
            return time;
        }

        /**
         * Returns the padding factor of the collection.
         *
         * @return the padding factor
         */
        public double getPaddingFactor() {
            return paddingFactor;
        }

        /**
         * Returns the size of the collection data in bytes.
         *
         * @return the size of the collection data in bytes
         */
        public long getDataSize() {
            return dataSize;
        }

        /**
         * Returns the space (extents) in bytes that is reserved for the collection data.
         *
         * @return the space in bytes
         */
        public long getDataStorageSize() {
            return dataStorageSize;
        }

        /**
         * The amount of documents inside the collection.
         *
         * @return the amount of documents
         */
        public long getDocumentCount() {
            return documentCount;
        }
    }

    /**
     * Resets the data of the profile.
     */
    void reset();

    /**
     * Returns the name of the collection.
     *
     * @return the name of the collection
     */
    String getCollectionName();

    /**
     * Determines whether the collection is capped. Returns {@code true} if the collection is capped (fixed size). Else
     * {@code false} is returned.
     *
     * @return {@code true} if the collection is capped, else {@code false}
     */
    boolean isCappedCollection();

    /**
     * Determines whether profiling data is available. Returns {@code true} if there is data available. Else {@code false}
     * is returned. If profiling data is available, {@link #getMarks()} must have at least one value.
     *
     * @return
     */
    boolean isDataAvailable();

    /**
     * Returns a Deque of marks (samples).
     *
     * @return the marks (samples)
     */
    Deque<Mark> getMarks();

    /**
     * Returns the measure of the average padding factor.
     *
     * @return the measure of the average padding factor
     */
    AverageMeasure getPaddingFactorMeasure();

    /**
     * Returns the measure of the average document count.
     *
     * @return the measure of the average document count
     */
    AverageMeasure getDocumentCountMeasure();

}
