/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.collection;

import de.otto.mongodb.profiler.AsyncProfiler;

import java.util.Collection;
import java.util.Set;

/**
 * A profiler that profiles a collection.
 *
 * @author Robert Gacki
 */
public interface CollectionProfiler extends AsyncProfiler {

    /**
     * Adds a new collection profile by the name of the collection. Throws a {@link CollectionDoesNotExistException} if
     * the collection is not found in the database. An implementation must not create a new collection. Adding
     * a collection that is already profiled just returns the {@link CollectionProfile}.
     *
     * @param collectionName the name of the collection
     * @return the profile for the collection
     * @throws CollectionDoesNotExistException
     *          if the collection does not exist
     */
    CollectionProfile addProfile(String collectionName) throws CollectionDoesNotExistException;

    /**
     * Removes a collection profile from the profiler. Does nothing if there is no collection profile found.
     *
     * @param collectionName the name of the collection
     */
    void removeProfile(String collectionName);

    /**
     * Returns all names of collections available to the profiler except for those that are already profiled and
     * system collections. Returns an empty set, if there are no more collections available for profiling.
     *
     * @return the set of collection names available for profiling or an empty set
     */
    Set<String> getAvailableCollections();

    /**
     * Returns all collection profiles. Returns an empty collection, if no collections are added to the profiler.
     *
     * @return the collection of profiles which may be empty
     */
    Collection<? extends CollectionProfile> getProfiles();

    /**
     * Returns a collection profile for a specific collection. Returns {@code null} if there is no profile
     * available from this profiler by the name.
     *
     * @param collectionName the name of the collection that is profiled
     * @return the collection profile or {@code null}
     */
    CollectionProfile getProfile(String collectionName);

    /**
     * Takes a new sample for each collection that is profiled. An implementation is able to decide whether the sample
     * is generated or to do nothing while the asynchronous profiling is active.
     */
    void newSample();

}
