/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Measures the movement of documents.
 *
 * @author Robert Gacki
 */
public class DocumentMoveMeasure {

    public static BigDecimal ratio(long normal, long moved) {
        if (moved == 0L) {
            return BigDecimal.valueOf(0.0);
        }
        final BigDecimal movedBd = BigDecimal.valueOf(moved);
        final BigDecimal total = BigDecimal.valueOf(normal).add(movedBd);
        return movedBd.divide(total, 5, RoundingMode.HALF_UP);
    }

    public static class Mark {

        public final long time;
        public final boolean moved;
        public final long normalCount;
        public final long movedCount;
        public final BigDecimal ratio;

        public Mark(long time, boolean moved, long normalCount, long movedCount) {
            this(time, moved, normalCount, movedCount, ratio(normalCount, movedCount));
        }

        public Mark(long time, boolean moved, long normalCount, long movedCount, BigDecimal ratio) {
            this.time = time;
            this.moved = moved;
            this.normalCount = normalCount;
            this.movedCount = movedCount;
            this.ratio = ratio;
        }
    }

    private final AtomicLong normalCount;
    private final AtomicLong movedCount;

    private final ConcurrentLinkedDeque<Mark> marks;

    public DocumentMoveMeasure() {
        this.normalCount = new AtomicLong();
        this.movedCount = new AtomicLong();
        this.marks = new ConcurrentLinkedDeque<>();
    }

    public void incrementMoved(long time) {
        final Mark mark = new Mark(time, true, normalCount.get(), movedCount.incrementAndGet());
        marks.add(mark);
    }

    public void incrementNormal(long time) {
        final Mark mark = new Mark(time, false, normalCount.incrementAndGet(), movedCount.get());
        marks.add(mark);
    }

    public BigDecimal getRatio() {
        return ratio(normalCount.get(), movedCount.get());
    }

    public long getMoved() {
        return movedCount.get();
    }

    public long getTotal() {
        return normalCount.get() + movedCount.get();
    }

    public Iterable<Mark> getMarks() {
        return marks;
    }

}
