/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op;

import de.otto.mongodb.profiler.AverageMeasure;

import java.util.Collection;
import java.util.List;

/**
 * A profile of an operation.
 *
 * @author Robert Gacki
 */
public interface OpProfile {

    /**
     * A configuration for dealing with outliers.
     *
     * @author Robert Gacki
     */
    public interface OutlierConfiguration {

        /**
         * Determines whether outliers are to be ignored for statistic calculations.
         *
         * @return
         */
        boolean ignoreOutliers();

        boolean captureOutliers();

        AverageMeasure.OutlierStrategy createStrategy();

    }

    /**
     * Returns the identity of the operation.
     *
     * @return the identity of the operation
     */
    String getId();

    /**
     * Returns the namespace, the operation is applied to.
     *
     * @return the namespace
     */
    String getNs();

    /**
     * Returns the operation type. An implementation should define a unique type for a certain type of operation in
     * order to support a frontend to differentiate between types.
     *
     * @return the operation type
     */
    String getType();

    /**
     * Returns the amount of events (operations) that occurred.
     *
     * @return the amount of events (operations)
     */
    long getEventCount();

    /**
     * Returns the last event (operation). Returns {@code null} if no event occured.
     *
     * @return the last event (operation) or {@code null}
     */
    ProfileEntry getLastEvent();

    /**
     * Returns the average measure of the execution time of the operation in milliseconds.
     *
     * @return the average measure of the execution time in milliseconds
     */
    AverageMeasure getAverageMillis();

    /**
     * Returns a collection of outliers of the average execution time in milliseconds.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageMillisOutliers() throws UnsupportedOperationException;

    /**
     * Returns the average measure of the time of the operation held a read lock in microseconds.
     *
     * @return the average measure of the time of the operation held a read lock in microseconds
     */
    AverageMeasure getAverageLockedReadMicros();

    /**
     * Returns a collection of outliers of the average read lock time in microseconds.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageLockedReadMicrosOutliers() throws UnsupportedOperationException;

    /**
     * Returns the average measure of the time of the operation held a write lock in microseconds.
     *
     * @return the average measure of the time of the operation held a write lock in microseconds
     */
    AverageMeasure getAverageLockedWriteMicros();

    /**
     * Returns a collection of outliers of the average write lock time in microseconds.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageLockedWriteMicrosOutliers() throws UnsupportedOperationException;

    /**
     * Returns the average measure of the time of the operation required to acquire a read lock in microseconds.
     *
     * @return the average measure of the time of the operation required to acquire a read lock in microseconds
     */
    AverageMeasure getAverageAcquireReadLockMicros();

    /**
     * Returns a collection of outliers of the average read lock acquisition time in microseconds.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageAcquireReadLockMicrosOutliers() throws UnsupportedOperationException;

    /**
     * Returns the average measure of the time of the operation required to acquire a write lock in microseconds.
     *
     * @return the average measure of the time of the operation required to acquire a write lock in microseconds
     */
    AverageMeasure getAverageAcquireWriteLockMicros();

    /**
     * Returns a collection of outliers of the average write lock acquisition time in microseconds.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageAcquireWriteLockMicrosOutliers() throws UnsupportedOperationException;

    /**
     * Returns the average measure of the amount of yields per operation.
     *
     * @return the average measure of the amount of yields per operation
     */
    AverageMeasure getAverageYieldCount();

    /**
     * Return the collection of outliers of the average measure of the amount of yields per operation.
     *
     * @return a collection of outliers
     * @throws UnsupportedOperationException if the profile does not support to capture outliers
     */
    Collection<AverageMeasure.Outlier> getAverageYieldCountOutliers();

    /**
     * Returns issues that were found. Returns an empty list if no issues were found yet.
     *
     * @return a list of issues or {@code null}
     */
    List<ProfileIssue> getIssues();

    /**
     * Dismisses an issue by its code. Has no affect if there is no issue by that code.
     *
     * @param code the code of the issue to dismiss
     */
    void dismissIssue(String code);

    /**
     * Returns the outlier configuration that is used by this profile to detect and handle outlying values. Returns
     * {@code null} if the profile does not handle outliers.
     *
     * @return the configuration or {@code null}
     */
    OutlierConfiguration getOutlierConfiguration();

}
