/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op;


import com.mongodb.DB;
import com.mongodb.DBCursor;
import com.mongodb.DBObject;
import de.otto.mongodb.profiler.op.analyze.Analyzer;
import de.otto.mongodb.profiler.op.analyze.AnalyzerContext;
import de.otto.mongodb.profiler.util.Logger;
import org.joda.time.DateTime;

import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.Collator;
import java.util.*;
import java.util.concurrent.atomic.AtomicReference;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static de.otto.mongodb.profiler.util.MongoSupport.get;

/**
 * The analyzer of the profiling data. Takes the data from a queue and passes them to the actual Analyzers.
 *
 * @author Robert Gacki
 */
class OpProfileAnalyzer implements Runnable {

    private final Logger logger = Logger.getLogger(OpProfileAnalyzer.class);

    private final Queue<DBObject> rawDataQueue;
    private final AnalyzerContextImpl analyzerContext;
    private final Collection<? extends Analyzer> analyzers;

    public OpProfileAnalyzer(final DB db,
                             final Queue<DBObject> rawDataQueue,
                             final Map<String, OpProfile> profiles,
                             final AtomicReference<DateTime> explanationMark,
                             final AtomicReference<OpProfile.OutlierConfiguration> outlierConfiguration,
                             final Collection<? extends Analyzer> analyzers) {
        this.rawDataQueue = checkNotNull(rawDataQueue);
        this.analyzerContext =
                new AnalyzerContextImpl(
                        checkNotNull(db),
                        checkNotNull(profiles),
                        checkNotNull(explanationMark),
                        checkNotNull(outlierConfiguration));
        this.analyzers = checkNotNull(analyzers);
    }

    @Override
    public synchronized void run() {

        try {
            DBObject dbo;
            while ((dbo = rawDataQueue.poll()) != null) {
                try {
                    handleOp(dbo);
                } catch (RuntimeException e) {
                    logger.error(e, "Failed to create profile entry!");
                }
            }
        } catch (RuntimeException e) {
            logger.error(e, "Failed to run!");
        }
    }

    protected void handleOp(DBObject dbo) {

        final Iterator<? extends Analyzer> analyzers = this.analyzers.iterator();
        while (analyzers.hasNext()) {
            final Analyzer analyzer = analyzers.next();
            if (analyzer.analyze(dbo, analyzerContext)) {
                logger.trace("Operation [%s] analyzed by [%s].", dbo, analyzer);
                return;
            }
        }

        logger.trace("Operation [%s] was not analyzed.", dbo);
    }

    private static class AnalyzerContextImpl implements AnalyzerContext {

        private static final Charset CHARSET = Charset.forName("utf-8");
        private static final Collator KEY_COLLATOR = Collator.getInstance(Locale.ENGLISH);

        private final DB db;
        private final Map<String, OpProfile> profiles;
        private final AtomicReference<DateTime> explanationMark;
        private final AtomicReference<OpProfile.OutlierConfiguration> outlierConfiguration;

        private AnalyzerContextImpl(final DB db,
                                    final Map<String, OpProfile> profiles,
                                    final AtomicReference<DateTime> explanationMark,
                                    final AtomicReference<OpProfile.OutlierConfiguration> outlierConfiguration) {
            this.db = checkNotNull(db);
            this.profiles = checkNotNull(profiles);
            this.explanationMark = checkNotNull(explanationMark);
            this.outlierConfiguration = checkNotNull(outlierConfiguration);
        }

        @Override
        public <T extends OpProfile> T findProfile(final String hash, final Class<T> profileType) throws ClassCastException {

            final OpProfile profile = profiles.get(hash);

            if (profile == null) {
                return null;
            }

            if (profileType.isInstance(profile) == false) {
                throw new ClassCastException(String.format(
                        "Invalid profile type [%s] for hash [%s]. Expected type is [%s]!",
                        profile.getClass(), hash, profileType));
            }

            return profileType.cast(profile);
        }

        @Override
        public <T extends OpProfile> T addProfile(final String hash, final T profile) {

            synchronized (profiles) {
                if (profiles.containsKey(hash)) {
                    throw new IllegalArgumentException(String.format("A query profile replaced another by hash [%s]! " +
                            "The context must not be accessed concurrently!", hash));
                }

                profiles.put(hash, profile);
            }

            return profile;
        }

        @Override
        public OpProfile.OutlierConfiguration getOutlierConfiguration() {
            return outlierConfiguration.get();
        }

        @Override
        public boolean shouldExplainQuery(final QueryProfile profile) {
            final Explanation explanation = profile.getLastExplanation();
            final DateTime mark = explanationMark.get();
            return explanation == null || mark.isAfter(explanation.getCreatedAt());
        }

        @Override
        public Explanation explainQuery(QueryProfile profile) {
            return explainQuery(profile.getNs(), profile);
        }

        @Override
        public Explanation explainQuery(final String ns, final QueryProfile profile) {

            checkNotNull(ns);
            checkNotNull(profile);

            final String collectionName = ns.substring(ns.indexOf('.') + 1);

            if (!db.collectionExists(collectionName)) {
                return null;
            }

            DBObject query = get("$query", DBObject.class, profile.getQuery());
            DBObject orderBy = null;

            if (query != null) {
                orderBy = get("$orderby", DBObject.class, profile.getQuery());
            } else {
                query = profile.getQuery();
            }

            try (DBCursor cursor = db.getCollection(collectionName).find(query)) {
                if (orderBy != null) {
                    cursor.sort(orderBy);
                }
                final DBObject explanationResult = cursor.explain();
                final Explanation explanation = new Explanation(explanationResult, profile, DateTime.now());
                return explanation;
            }
        }

        @Override
        public String createQueryHash(final String ns, final String op, final DBObject query) {
            try {
                final MessageDigest digest = MessageDigest.getInstance("sha-1");
                digest.update(ns.getBytes(CHARSET));
                digest.update(op.getBytes(CHARSET));
                update(digest, query);
                return "query-" + byteToHex(digest.digest());
            } catch (NoSuchAlgorithmException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public String createMapReduceHash(String ns, String mapFunction, String reduceFunction) {
            try {
                final MessageDigest digest = MessageDigest.getInstance("sha-1");
                digest.update(ns.getBytes(CHARSET));
                digest.update(mapFunction.getBytes(CHARSET));
                digest.update(reduceFunction.getBytes(CHARSET));
                return "mapreduce-" + byteToHex(digest.digest());
            } catch (NoSuchAlgorithmException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public String createAggregateHash(String ns, List<DBObject> pipeline) {

            checkArgument(!checkNotNull(pipeline).isEmpty());

            try {
                final MessageDigest digest = MessageDigest.getInstance("sha-1");
                digest.update(ns.getBytes(CHARSET));
                for (DBObject dbo : pipeline) {
                    update(digest, dbo);
                }
                return "aggregate-" + byteToHex(digest.digest());
            } catch (NoSuchAlgorithmException e) {
                throw new RuntimeException(e);
            }
        }

        private static void update(final MessageDigest digest, final DBObject dbo) {
            final List<String> keys = new ArrayList<>(dbo.keySet());
            Collections.sort(keys, KEY_COLLATOR);
            for (String key : keys) {
                digest.update(key.getBytes(CHARSET));
                final Object value = dbo.get(key);
                if (value instanceof DBObject) {
                    update(digest, ((DBObject) value));
                }
            }
        }

        private static String byteToHex(final byte[] bytes) {

            final StringBuilder sb = new StringBuilder(bytes.length * 2);
            int inv;
            for (byte aByte : bytes) {
                inv = aByte & 0xff;
                if (inv < 16) {
                    sb.append(0);
                }
                sb.append(Integer.toHexString(inv));
            }
            return sb.toString();
        }
    }

}
