/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op;

import de.otto.mongodb.profiler.AsyncProfiler;

import java.util.Collection;

/**
 * <p>A profiler that analyzes MongoDB operations. The profiler must be able to modify the profiling level of the database.
 * The profiling should be done asynchronously.</p>
 * <p/>
 * <p>The profiler must not analyze operations of system collections, especially the profiling collection itself.</p>
 *
 * @author Robert Gacki
 */
public interface OpProfiler extends AsyncProfiler {

    /**
     * The profiling level of the MongoDB (database).
     *
     * @author Robert Gacki
     */
    public static enum ProfilingLevel {

        /**
         * Profiling is deactivated.
         */
        OFF(0),

        /**
         * Only slow operations are profiled. Please refer to the MongoDB documentation and your setup to see which
         * operations are affected by this level.
         */
        SLOW_OPS(1),

        /**
         * Level that profiles all operations.
         */
        ALL_OPS(2);

        public final int value;

        private ProfilingLevel(int value) {
            this.value = value;
        }

        public static ProfilingLevel forValue(int value) {
            for (ProfilingLevel level : values()) {
                if (value == level.value) {
                    return level;
                }
            }
            return null;
        }

        public int getValue() {
            return value;
        }
    }

    /**
     * Requests and returns the current profiling level of the MongoDB instance or database. Must throw a
     * RuntimeException if the profiling level could not be determined.
     *
     * @return the current profiling level of the MongoDB instance or database
     * @throws RuntimeException if the profiling level could not be determined or for other runtime related issues
     */
    ProfilingLevel getProfilingLevel() throws RuntimeException;

    /**
     * Sets the profiling level of the MongoDB instance or database. Must throw a
     * RuntimeException if the profiling level could not be set.
     *
     * @param level the profiling level to set
     * @throws RuntimeException if the profiling level could not be set or for other runtime related issues
     */
    void setProfilingLevel(ProfilingLevel level) throws RuntimeException;

    /**
     * Returns all query profiles.
     *
     * @return the query profiles
     * @deprecated use {@link #getProfiles()} instead
     */
    @Deprecated
    Collection<? extends QueryProfile> getQueryProfiles();

    /**
     * Returns a query profile by its identity. Returns {@code null} if there is no such profile.
     *
     * @param id the identity of the profile
     * @return the profile or {@code null} if not found
     * @deprecated use {@link #getProfile(String)} instead
     */
    @Deprecated
    QueryProfile getQueryProfile(String id);

    /**
     * Returns all query profiles.
     *
     * @return the query profiles
     */
    Collection<? extends OpProfile> getProfiles();

    /**
     * Returns a query profile by its identity. Returns {@code null} if there is no such profile.
     *
     * @param id the identity of the profile
     * @return the profile or {@code null} if not found
     */
    OpProfile getProfile(String id);

    /**
     * Forces the profiler to explain the next queries of each query identity.
     */
    void explainNextQueries();

    /**
     * Returns the outlier configuration for the profiles. Returns {@code null} if outliers are not handled. If set,
     * profiles must use this configuration to identify and handle outlying values.
     *
     * @return the outlier configuration or {@code null}
     */
    OpProfile.OutlierConfiguration getOutlierConfiguration();

    /**
     * Sets the outlier configuration for the profiles. If set to {@code null}, outliers will not be dealt with. If a
     * configuration is applied, the implementation may reset the profiling data in order to apply new calculation
     * rules.
     *
     * @param configuration the configuration to set which may be {@code null} in order to reset outlier handling
     */
    void setOutlierConfiguration(OpProfile.OutlierConfiguration configuration);

}
