/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op;

import java.util.concurrent.atomic.AtomicLong;

/**
 * An issue related to a profile.
 *
 * @author Robert Gacki
 */
public class ProfileIssue {

    public static enum Severity {
        WARNING;
    }

    private final String code;
    private final Severity severity;
    private final AtomicLong occurrences;

    public ProfileIssue(final String code,
                        final Severity severity) {
        this.code = code;
        this.severity = severity;
        this.occurrences = new AtomicLong(1L);
    }

    /**
     * Returns the code of the issue that makes it distinguishable to other issues.
     *
     * @return the code
     */
    public String getCode() {
        return code;
    }

    /**
     * Returns the severity of the issue.
     *
     * @return the severity
     */
    public Severity getSeverity() {
        return severity;
    }

    /**
     * Adds a new occurrence to this issue and returns the issue instance.
     *
     * @return this instance
     */
    public ProfileIssue newOccurrence() {
        occurrences.incrementAndGet();
        return this;
    }

    /**
     * Returns the amount of occurrences for this issue.
     *
     * @return the amount of occurrences
     */
    public long getOccurrences() {
        return occurrences.get();
    }

}
