/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op.analyze;

import com.mongodb.DBObject;
import de.otto.mongodb.profiler.op.Explanation;
import de.otto.mongodb.profiler.op.OpProfile;
import de.otto.mongodb.profiler.op.QueryProfile;

import java.util.List;

/**
 * A context provided to an analyzer for profile management.
 *
 * @author Robert Gacki
 */
public interface AnalyzerContext {

    /**
     * Finds a profile of the hashed operation and a certain profile type. Returns {@code null} if a profile by this
     * hash was not found. Throws a ClassCastException if the type of the profile does not match the provided
     * {@code profileType}.
     *
     * @param hash        the hash of the operation
     * @param profileType the class of the profile type
     * @param <T>         the profile type
     * @return the profile or {@code null} if not found
     * @throws ClassCastException if the type of the profile found does not match the requested type
     */
    <T extends OpProfile> T findProfile(String hash, Class<T> profileType) throws ClassCastException;

    /**
     * Adds a profile to the known profiles using a hash. Throws an IllegalArgumentException if that hash is already
     * added. Returns the profile, when it is added.
     *
     * @param hash    the hash of the profile
     * @param profile the profile to add
     * @param <T>     the profile type
     * @return the added profile
     */
    <T extends OpProfile> T addProfile(String hash, T profile) throws IllegalArgumentException;

    /**
     * Creates a hash for a operation that has a query as context.
     *
     * @param ns    the namespace of the operation
     * @param op    the operation name
     * @param query the query
     * @return the hash
     */
    String createQueryHash(String ns, String op, DBObject query);

    /**
     * Creates a hash for a map-reduce operation.
     *
     * @param ns             the namespace of the operation
     * @param mapFunction    the map function
     * @param reduceFunction the reduce function
     * @return the hash
     */
    String createMapReduceHash(String ns, String mapFunction, String reduceFunction);

    /**
     * Creates a hash for an aggregate operation.
     *
     * @param ns       the namespace of the operation
     * @param pipeline the pipeline of the aggregation
     * @return the hash
     */
    String createAggregateHash(String ns, List<DBObject> pipeline);

    /**
     * Determines whether a query should be explained. Returns {@code true} if the query should be explained, else
     * {@code false}.
     *
     * @param profile the query profile to check
     * @return {@code true} if the query should be explained, else {@code false}
     */
    boolean shouldExplainQuery(QueryProfile profile);

    /**
     * Explains the query of a query profile.
     *
     * @param profile the query profile to explain
     * @return
     */
    Explanation explainQuery(QueryProfile profile);

    /**
     * Explains the query of a query profile.
     *
     * @param ns      the namespace of the query
     * @param profile the query profile to explain
     * @return
     */
    Explanation explainQuery(String ns, QueryProfile profile);

    /**
     * Returns the outlier configuration or {@code null} if none is set. If set, the analyzer should apply the
     * configuration to a profile in order to detect outlying samples.
     *
     * @return the outlier configuration or {@code null} if none is set
     */
    OpProfile.OutlierConfiguration getOutlierConfiguration();

}
