/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op.analyze;

import com.mongodb.DBObject;
import de.otto.mongodb.profiler.AverageMeasure;
import de.otto.mongodb.profiler.op.OpProfile;
import de.otto.mongodb.profiler.op.ProfileEntry;
import de.otto.mongodb.profiler.op.ProfileIssue;

import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;

import static com.google.common.base.Preconditions.checkNotNull;
import static de.otto.mongodb.profiler.util.MongoSupport.get;

/**
 * @author Robert Gacki
 */
abstract class DefaultOpProfile implements OpProfile {

    private static AverageMeasure createMeasure(final OutlierConfiguration config,
                                                final Collection<AverageMeasure.Outlier> outliers) {

        if (config == null) {
            return new AverageMeasure();
        }

        final AverageMeasure.OutlierStrategy strategy;
        if (outliers != null) {
            strategy = new AverageMeasure.CapturingOutlierStrategy(config.createStrategy(), outliers);
        } else {
            strategy = config.createStrategy();
        }

        return new AverageMeasure(strategy, config.ignoreOutliers());
    }

    public static class MeasureAndOutliers {

        private volatile AverageMeasure measure;
        private final Collection<AverageMeasure.Outlier> outliers;

        private MeasureAndOutliers(final OutlierConfiguration config) {
            this.outliers = config != null && config.captureOutliers() ? new ConcurrentLinkedQueue<AverageMeasure.Outlier>() : null;
            this.measure = createMeasure(config, this.outliers);
        }

        public void add(double amount) {
            measure = measure.add(amount);
        }

        private AverageMeasure getMeasure() {
            return measure;
        }

        private Collection<AverageMeasure.Outlier> getOutliers() {
            return outliers;
        }
    }

    private final String id;
    private final String ns;
    private final AtomicLong eventCount;
    private final AtomicReference<ProfileEntry> lastEvent;

    private final Map<String, ProfileIssue> issues;

    private final OutlierConfiguration outlierConfiguration;

    private final MeasureAndOutliers averageMillis;
    private final MeasureAndOutliers averageLockedReadMicros;
    private final MeasureAndOutliers averageLockedWriteMicros;
    private final MeasureAndOutliers averageAcquireReadLockMicros;
    private final MeasureAndOutliers averageAcquireWriteLockMicros;
    private final MeasureAndOutliers averageYieldCount;

    public DefaultOpProfile(final String id,
                            final String ns,
                            final OutlierConfiguration outlierConfiguration) {
        this.id = id;
        this.ns = ns;
        this.eventCount = new AtomicLong(0L);
        this.lastEvent = new AtomicReference<>();
        this.issues = Collections.synchronizedMap(new LinkedHashMap<String, ProfileIssue>());

        this.outlierConfiguration = outlierConfiguration;

        this.averageMillis = new MeasureAndOutliers(outlierConfiguration);
        this.averageLockedReadMicros = new MeasureAndOutliers(outlierConfiguration);
        this.averageLockedWriteMicros = new MeasureAndOutliers(outlierConfiguration);
        this.averageAcquireReadLockMicros = new MeasureAndOutliers(outlierConfiguration);
        this.averageAcquireWriteLockMicros = new MeasureAndOutliers(outlierConfiguration);
        this.averageYieldCount = new MeasureAndOutliers(outlierConfiguration);
    }

    @Override
    public String getId() {
        return id;
    }

    @Override
    public String getNs() {
        return ns;
    }

    @Override
    public long getEventCount() {
        return this.eventCount.get();
    }

    private static void add(MeasureAndOutliers mao, Number number) {
        if (number != null) {
            mao.add(number.doubleValue());
        }
    }

    public void add(ProfileEntry entry) {

        this.lastEvent.set(entry);
        this.eventCount.incrementAndGet();

        final DBObject data = entry.getData();

        add(this.averageMillis, get("millis", Number.class, data));

        final DBObject lockStats = get("lockStats", DBObject.class, data);

        if (lockStats != null) {
            final DBObject lockedMicros = get("timeLockedMicros", DBObject.class, lockStats);
            if (lockedMicros != null) {
                add(this.averageLockedReadMicros, get("r", Number.class, lockedMicros));
                add(this.averageLockedWriteMicros, get("w", Number.class, lockedMicros));
            }

            final DBObject acquiringMicros = get("timeAcquiringMicros", DBObject.class, lockStats);
            if (acquiringMicros != null) {
                add(this.averageAcquireReadLockMicros, get("r", Number.class, acquiringMicros));
                add(this.averageAcquireWriteLockMicros, get("w", Number.class, acquiringMicros));
            }
        }

        final Number yields = get("numYield", Number.class, data);
        if (yields != null) {
            add(this.averageYieldCount, yields.doubleValue());
        }

        final Collection<ProfileIssue> issues = findIssuesInProfile(entry.getData());
        if (issues != null && !issues.isEmpty()) {
            synchronized (this.issues) {
                for (ProfileIssue issue : issues) {
                    final ProfileIssue existing = this.issues.get(issue.getCode());
                    if (existing != null) {
                        existing.newOccurrence();
                    } else {
                        this.issues.put(issue.getCode(), issue);
                    }
                }
            }
        }
    }

    protected Collection<ProfileIssue> findIssuesInProfile(DBObject profile) {

        final List<ProfileIssue> result = new ArrayList<>(1);
        return result;
    }

    @Override
    public ProfileEntry getLastEvent() {
        return lastEvent.get();
    }

    @Override
    public AverageMeasure getAverageMillis() {
        return averageMillis.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageMillisOutliers() throws UnsupportedOperationException {
        return averageMillis.getOutliers();
    }

    @Override
    public AverageMeasure getAverageLockedReadMicros() {
        return averageLockedReadMicros.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageLockedReadMicrosOutliers() throws UnsupportedOperationException {
        return averageLockedReadMicros.getOutliers();
    }

    @Override
    public AverageMeasure getAverageLockedWriteMicros() {
        return averageLockedWriteMicros.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageLockedWriteMicrosOutliers() throws UnsupportedOperationException {
        return averageLockedWriteMicros.getOutliers();
    }

    @Override
    public AverageMeasure getAverageAcquireReadLockMicros() {
        return averageAcquireReadLockMicros.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageAcquireReadLockMicrosOutliers() throws UnsupportedOperationException {
        return averageAcquireReadLockMicros.getOutliers();
    }

    @Override
    public AverageMeasure getAverageAcquireWriteLockMicros() {
        return averageAcquireWriteLockMicros.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageAcquireWriteLockMicrosOutliers() throws UnsupportedOperationException {
        return averageAcquireWriteLockMicros.getOutliers();
    }

    @Override
    public AverageMeasure getAverageYieldCount() {
        return averageYieldCount.getMeasure();
    }

    @Override
    public Collection<AverageMeasure.Outlier> getAverageYieldCountOutliers() {
        return averageYieldCount.getOutliers();
    }

    @Override
    public List<ProfileIssue> getIssues() {
        return new ArrayList<>(issues.values());
    }

    @Override
    public void dismissIssue(String code) {
        this.issues.remove(checkNotNull(code));
    }

    @Override
    public OutlierConfiguration getOutlierConfiguration() {
        return outlierConfiguration;
    }
}
