/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op.analyze;

import com.mongodb.DBObject;
import de.otto.mongodb.profiler.op.Explanation;
import de.otto.mongodb.profiler.op.ProfileIssue;
import de.otto.mongodb.profiler.op.QueryProfile;

import java.util.Collection;
import java.util.Deque;
import java.util.NoSuchElementException;
import java.util.concurrent.ConcurrentLinkedDeque;

/**
 * @author Robert Gacki
 */
class DefaultQueryProfile extends DefaultOpProfile implements QueryProfile {

    public static enum QueryProfileIssues {

        DOCS_SCANNED_DIFFERS_FROM_RETURNED(ProfileIssue.Severity.WARNING);

        public final ProfileIssue issue;

        private QueryProfileIssues(final ProfileIssue.Severity severity) {
            issue = new ProfileIssue(getClass().getSimpleName() + "." + name(), severity);
        }
    }

    private final DBObject query;
    private final Deque<Explanation> explanations;
    private final boolean snapshot;

    public DefaultQueryProfile(final String id,
                               final String ns,
                               final OutlierConfiguration outlierConfiguration,
                               final DBObject query,
                               final boolean snapshot) {
        super(id, ns, outlierConfiguration);
        this.query = query;
        this.explanations = new ConcurrentLinkedDeque<>();
        this.snapshot = snapshot;
    }

    @Override
    public DBObject getQuery() {
        return query;
    }

    @Override
    public boolean isSnapshot() {
        return snapshot;
    }

    @Override
    public String getType() {
        return TYPE_VALUE;
    }

    @Override
    public int getExplanationCount() {
        return this.explanations.size();
    }

    @Override
    public Iterable<Explanation> getExplanations() {
        return explanations;
    }

    @Override
    public Explanation getLastExplanation() {
        try {
            return explanations.getFirst();
        } catch (NoSuchElementException e) {
            return null;
        }
    }

    public void add(Explanation explanation) {
        explanations.push(explanation);
    }

    @Override
    protected Collection<ProfileIssue> findIssuesInProfile(DBObject profile) {

        final Collection<ProfileIssue> issues = super.findIssuesInProfile(profile);

        final Number scanned = ((Number) profile.get("nscanned"));
        final Number returned = ((Number) profile.get("nreturned"));
        if (scanned != null && returned != null && scanned.longValue() > returned.longValue()) {
            issues.add(QueryProfileIssues.DOCS_SCANNED_DIFFERS_FROM_RETURNED.issue);
        }

        return issues;
    }
}
