/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.op.analyze;

import com.mongodb.DBObject;
import de.otto.mongodb.profiler.op.Explanation;
import de.otto.mongodb.profiler.op.GetMoreProfile;
import de.otto.mongodb.profiler.op.OpProfile;
import de.otto.mongodb.profiler.op.ProfileEntry;
import org.joda.time.DateTime;

import static de.otto.mongodb.profiler.util.MongoSupport.get;
import static de.otto.mongodb.profiler.util.MongoSupport.require;

/**
 * An analyzer of 'simple' queries which are 'find', 'getmore' and 'update' operations.
 *
 * @author Robert Gacki
 */
public class SimpleQueryAnalyzer implements Analyzer {

    private static enum QueryType {

        QUERY("query", DefaultQueryProfile.class, true) {
            @Override
            public DefaultQueryProfile createProfile(String queryHash,
                                                     String ns,
                                                     OpProfile.OutlierConfiguration outlierConfiguration,
                                                     DBObject queryData) {
                final boolean snapshotQuery = queryData.containsField("$snapshot")
                        && Boolean.TRUE.equals(get("$snapshot", Boolean.class, queryData, Boolean.FALSE));
                return new DefaultQueryProfile(queryHash, ns, outlierConfiguration, queryData, snapshotQuery);
            }
        },
        GET_MORE("getmore", DefaultGetMoreProfile.class, false) {
            @Override
            public DefaultQueryProfile createProfile(String queryHash,
                                                     String ns,
                                                     OpProfile.OutlierConfiguration outlierConfiguration,
                                                     DBObject queryData) {
                return new DefaultGetMoreProfile(queryHash, ns, outlierConfiguration, queryData);
            }
        },
        UPDATE("update", DefaultUpdateProfile.class, true) {
            @Override
            public DefaultQueryProfile createProfile(String queryHash,
                                                     String ns,
                                                     OpProfile.OutlierConfiguration outlierConfiguration,
                                                     DBObject queryData) {
                return new DefaultUpdateProfile(queryHash, ns, outlierConfiguration, queryData);
            }
        };

        public final String operation;
        public final Class<? extends DefaultQueryProfile> profileType;
        public final boolean explain;

        private QueryType(final String operation,
                          final Class<? extends DefaultQueryProfile> profileType,
                          final boolean explain) {
            this.operation = operation;
            this.profileType = profileType;
            this.explain = explain;
        }

        public abstract DefaultQueryProfile createProfile(String queryHash,
                                                          String ns,
                                                          OpProfile.OutlierConfiguration outlierConfiguration,
                                                          DBObject queryData);

        public static QueryType typeOf(DBObject dbo) {

            final String op = require("op", String.class, dbo);

            if (op != null) {
                for (QueryType type : QueryType.values()) {
                    if (type.operation.equals(op)) {
                        return type;
                    }
                }
            }

            return null;
        }
    }

    @Override
    public boolean analyze(DBObject dbo, AnalyzerContext context) {

        final QueryType type = QueryType.typeOf(dbo);

        if (type == null) {
            return false;
        }

        final DBObject queryData = require("query", DBObject.class, dbo);

        if (get("$explain", Boolean.class, queryData, Boolean.FALSE).booleanValue()) {
            return false;
        }

        final String op = require("op", String.class, dbo);
        final String ns = require("ns", String.class, dbo);
        final String hash = context.createQueryHash(ns, op, queryData);

        DefaultQueryProfile profile = context.findProfile(hash, type.profileType);

        if (profile == null) {
            profile = type.createProfile(hash, ns, context.getOutlierConfiguration(), queryData);
            profile = context.addProfile(hash, profile);
        }

        if (type.explain && context.shouldExplainQuery(profile)) {
            final Explanation explanation = context.explainQuery(profile);
            if (explanation != null) {
                profile.add(explanation);
            }
        }

        profile.add(new ProfileEntry(dbo, DateTime.now()));

        return true;
    }

    private static class DefaultGetMoreProfile extends DefaultQueryProfile implements GetMoreProfile {

        public DefaultGetMoreProfile(String id,
                                     String ns,
                                     OutlierConfiguration outlierConfiguration,
                                     DBObject query) {
            super(id, ns, outlierConfiguration, query, false);
        }

        @Override
        public String getType() {
            return GetMoreProfile.TYPE_VALUE;
        }
    }

}
