/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.util;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * Utility class to support asynchronous operations.
 *
 * @author Robert Gacki
 */
public class Async {

    /**
     * Tries to shut down an ExecutorService gracefully. Waits for tasks to finish for a specific amount of 'waitTime'. If the
     * executor did not shut down in time, {@link java.util.concurrent.ExecutorService#shutdownNow()} is executed and
     * another wait occurs. Worst case is that the call will throw an IllegalStateException stating that the executor
     * service was not shut down in time. That time may then be two-times of the 'waitTime'.
     *
     * @param service      the executor service to shut down
     * @param waitTime     the time to wait for the executor service to be shut down (worst case is two-times of that time)
     * @param waitTimeUnit the time unit to wait for the executor service to be shut down
     * @see java.util.concurrent.ExecutorService#shutdown()
     * @see java.util.concurrent.ExecutorService#shutdownNow()
     * @see java.util.concurrent.ExecutorService#awaitTermination(long, java.util.concurrent.TimeUnit)
     */
    public static void shutdownAndAwaitTermination(final ExecutorService service,
                                                   final int waitTime,
                                                   final TimeUnit waitTimeUnit) {
        service.shutdown();
        try {
            if (!service.awaitTermination(waitTime, waitTimeUnit)) {
                service.shutdownNow();
                if (!service.awaitTermination(waitTime, waitTimeUnit)) {
                    throw new IllegalStateException(
                            String.format("Executor service [%s] did not shut down in time!", service));
                }
            }
        } catch (InterruptedException e) {
            service.shutdownNow();
            Thread.currentThread().interrupt();
        }
    }
}
