/*
 *  Copyright 2013 Robert Gacki <robert.gacki@cgi.com>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package de.otto.mongodb.profiler.util;

import com.mongodb.DBObject;

import static java.lang.String.format;

/**
 * Utility methods for MongoDB.
 *
 * @author Robert Gacki
 */
public class MongoSupport {

    /**
     * Requires a DBObject to have a specific field of a specific type. Returns the value of that field if available.
     * Throws and IllegalArgumentException if either the value does not exist or is not of the designated type.
     *
     * @param field  the name of the field
     * @param asType the expected class of the type
     * @param from   the DBObject to return the field value from
     * @param <T>    the expected type
     * @return the field value cast to the expected type
     */
    public static <T> T require(String field, Class<T> asType, DBObject from) {

        final T value = get(field, asType, from);

        if (value == null) {
            throw new IllegalArgumentException(format("No value of type [%s] available for field [%s]! " +
                    "See debug message for details.", asType, field));
        }

        return value;
    }

    /**
     * Returns the value of a specific field and type if available. Returns {@code null} if either the value does not
     * exist or is not of the designated type.
     *
     * @param field  the name of the field
     * @param asType the expected class of the type
     * @param from   the DBObject to return the field value from
     * @param <T>    the expected type
     * @return the field value cast to the expected type
     */
    public static <T> T get(String field, Class<T> asType, DBObject from) {

        return get(field, asType, from, null);
    }

    /**
     * Returns the value of a specific field and type if available. Returns 'elseReturn' if either the value does not
     * exist or is not of the designated type.
     *
     * @param field      the name of the field
     * @param asType     the expected class of the type
     * @param from       the DBObject to return the field value from
     * @param elseReturn the value to return if the DBObject does either not contain the field or has no value matching the expected type
     * @param <T>        the expected type
     * @return the field value cast to the expected type
     */
    public static <T> T get(String field, Class<T> asType, DBObject from, T elseReturn) {

        final Object value = from.get(field);

        if (value == null) {
            return elseReturn;
        }

        if (!asType.isInstance(value)) {
            throw new IllegalArgumentException(format("Value <<censored>> is not of type [%s] as expected " +
                    "for field [%s]! See debug message for details.", asType, field));
        }

        return ((T) value);
    }

}
