package de.otto.mongodb.profiler;

import org.testng.annotations.Test;

import java.math.BigDecimal;
import java.math.MathContext;
import java.util.ArrayList;
import java.util.List;

import static org.testng.Assert.*;

public class AverageMeasureTest extends FunctionTest {

    @Test
    public void shouldAddValueToDefaultStale() {

        // Given
        final AverageMeasure measure = new AverageMeasure();

        // When
        final AverageMeasure result = measure.add(7.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getCount(), 1L);
        assertEquals(result.getAverage().compareTo(BigDecimal.valueOf(7.0)), 0);
        assertEquals(result.getMin(), 7.0);
        assertEquals(result.getMax(), 7.0);
    }

    @Test
    public void shouldAddValueToDefaultNonStale() {

        // Given
        final AverageMeasure measure = new AverageMeasure(false);

        // When
        final AverageMeasure result = measure.add(7.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getCount(), 1L);
        assertEquals(result.getAverage().compareTo(BigDecimal.valueOf(7.0)), 0);
        assertEquals(result.getMin(), 0.0);
        assertEquals(result.getMax(), 7.0);
    }

    @Test
    public void shouldAddValueToMutated() {

        // Given
        final AverageMeasure measure = new AverageMeasure().add(7.0);
        assertEquals(measure.getCount(), 1L);
        assertEquals(measure.getAverage().compareTo(BigDecimal.valueOf(7.0)), 0);
        assertEquals(measure.getMin(), 7.0);
        assertEquals(measure.getMax(), 7.0);

        // When
        final AverageMeasure result = measure.add(7.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getAverage().compareTo(BigDecimal.valueOf(7.0)), 0);
        assertEquals(result.getCount(), 2L);
        assertEquals(result.getMin(), 7.0);
        assertEquals(result.getMax(), 7.0);
    }

    @Test
    public void shouldCalculateAverage() {

        // Given
        final AverageMeasure measure = new AverageMeasure().add(109.0).add(98.0);

        // When
        final BigDecimal average = measure.getAverage();

        // Then
        assertNotNull(average);
        assertEquals(average, BigDecimal.valueOf(103.5));
    }

    @Test
    public void shouldCalculateAverageWithMathContext() {

        // Given
        final AverageMeasure measure = new AverageMeasure().add(109.0).add(98.0);
        final MathContext mathContext = MathContext.DECIMAL64;

        // When
        final BigDecimal average = measure.getAverage(mathContext);

        // Then
        assertNotNull(average);
        assertEquals(average, BigDecimal.valueOf(103.5));
    }

    @Test
    public void shouldAddAbsoluteValues() {

        // Given
        final AverageMeasure measure = new AverageMeasure();

        // When
        final AverageMeasure result = measure.add(-7.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getCount(), 1L);
        assertEquals(result.getAverage().compareTo(BigDecimal.valueOf(7.0)), 0);
        assertEquals(result.getMin(), 7.0);
        assertEquals(result.getMax(), 7.0);
    }

    @Test(expectedExceptions = RuntimeException.class)
    public void shouldNotAddToMeasureWithMaximumCount() {

        // Given
        final AverageMeasure measure = new AverageMeasure(
                AverageMeasure.DEFAULT_VALUE,
                AverageMeasure.DEFAULT_MIN,
                AverageMeasure.DEFAULT_MAX,
                Long.MAX_VALUE);

        // When
        measure.add(0.0);

        // Then
    }

    @Test
    public void shouldCalculateMinValue() {

        // Given
        final AverageMeasure measure = new AverageMeasure(
                AverageMeasure.DEFAULT_VALUE,
                11.0,
                11.0,
                AverageMeasure.DEFAULT_COUNT);

        // When
        final AverageMeasure result = measure.add(7.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getMin(), 7.0);
        assertEquals(result.getMax(), 11.0);
    }

    @Test
    public void shouldCalculateMaxValue() {

        // Given
        final AverageMeasure measure = new AverageMeasure(
                AverageMeasure.DEFAULT_VALUE,
                11.0,
                11.0,
                AverageMeasure.DEFAULT_COUNT);

        // When
        final AverageMeasure result = measure.add(17.0);

        // Then
        assertNotSame(result, measure);
        assertEquals(result.getMin(), 11.0);
        assertEquals(result.getMax(), 17.0);
    }

    @Test
    public void shouldNotCountOutlierForTwoTimesOfTheMeanValueStrategy() {

        final double value = 11.0;
        final AverageMeasure measure = new AverageMeasure(AverageMeasure.TWO_TIMES_MEAN_VALUE, false).add(value);
        assertEquals(measure.getOutlierCount(), 0L);

        // When
        final AverageMeasure result = measure.add(value * 2);

        // Then
        assertEquals(result.getOutlierCount(), 0L);
    }

    @Test
    public void shouldCountOutlierForTwoTimesOfTheMeanValueStrategy() {

        final double value = 11.0;
        final AverageMeasure measure = new AverageMeasure(AverageMeasure.TWO_TIMES_MEAN_VALUE, false).add(value);
        assertEquals(measure.getOutlierCount(), 0L);

        // When
        final AverageMeasure result = measure.add(value * 2 + 0.1);

        // Then
        assertEquals(result.getOutlierCount(), 1L);
        assertNotEquals(result.getAverage().compareTo(BigDecimal.valueOf(value)), 0);
    }

    @Test
    public void shouldIgnoreOutlierValue() {

        final double value = 11.0;
        final double outlier = 77.0;
        final AverageMeasure.OutlierStrategy strategy = new AverageMeasure.OutlierStrategy() {

            @Override
            public boolean isOutlier(double value, AverageMeasure current) {
                return value == outlier;
            }
        };
        final AverageMeasure measure = new AverageMeasure(strategy, true).add(value);
        assertEquals(measure.getOutlierCount(), 0L);

        // When
        final AverageMeasure result = measure.add(outlier);

        // Then
        assertEquals(result.getOutlierCount(), 1L);
        assertEquals(result.getAverage().compareTo(BigDecimal.valueOf(value)), 0);
    }

    @Test
    public void shouldCaptureOutlier() {

        final double value = 11.0;
        final double outlierValue = 77.0;
        final List<AverageMeasure.Outlier> outliers = new ArrayList<>();
        final AverageMeasure.CapturingOutlierStrategy strategy = new AverageMeasure.CapturingOutlierStrategy(new AverageMeasure.OutlierStrategy() {

            @Override
            public boolean isOutlier(double value, AverageMeasure current) {
                return value == outlierValue;
            }
        }, outliers);
        final AverageMeasure measure = new AverageMeasure(strategy, false).add(value);
        assertEquals(measure.getOutlierCount(), 0L);
        assertEquals(outliers.size(), 0);

        // When
        final AverageMeasure result = measure.add(outlierValue);

        // Then
        assertEquals(result.getOutlierCount(), 1L);
        assertEquals(outliers.size(), 1);
        final AverageMeasure.Outlier outlier = outliers.iterator().next();
        assertEquals(outlier.getValue(), outlierValue);
        assertSame(outlier.getMeasure(), measure);
    }

}
