package de.otto.mongodb.profiler.collection;

import com.mongodb.CommandResult;
import com.mongodb.DB;
import com.mongodb.DBCollection;
import de.otto.mongodb.profiler.AbstractAsyncProfilerTest;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.concurrent.ScheduledExecutorService;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.*;
import static org.testng.Assert.*;

public class DefaultCollectionProfilerTest extends AbstractAsyncProfilerTest {

    private ScheduledExecutorService analyzerExecutorServiceMock;
    private ScheduledExecutorService managementExecutorServiceMock;
    private DB dbMock;

    @Override
    protected DefaultCollectionProfiler createProfiler(boolean execute) {
        final DefaultCollectionProfiler profiler = new DefaultCollectionProfiler(
                execute, dbMock,
                analyzerExecutorServiceMock, managementExecutorServiceMock);
        return profiler;
    }

    @BeforeMethod
    public void setUp() {

        analyzerExecutorServiceMock = mock(ScheduledExecutorService.class);
        managementExecutorServiceMock = mock(ScheduledExecutorService.class);
        dbMock = mock(DB.class);
    }

    @Test
    public void shouldInitializeDelayedCollectionJob() {

        // Given
        final DefaultCollectionProfiler profiler = createProfiler(false);

        // When
        profiler.initialize();

        // Then
        verify(analyzerExecutorServiceMock).scheduleWithFixedDelay(
                any(DefaultCollectionProfiler.CollectDataJob.class),
                eq(DefaultCollectionProfiler.DEFAULT_INITIAL_DELAY),
                eq(DefaultCollectionProfiler.DEFAULT_DELAY),
                eq(DefaultCollectionProfiler.DEFAULT_TIMEUNIT));
    }

    @Test
    public void shouldShutdown() {

        // Given
        final DefaultCollectionProfiler profiler = createProfiler(false);
        immediateShutdown(analyzerExecutorServiceMock);

        // When
        profiler.shutdown();

        // Then
        verify(analyzerExecutorServiceMock).shutdown();
    }

    @Test
    public void shouldUpdateCollectionStatisticsAndReset() throws CollectionDoesNotExistException {

        // Given
        final String collectionName = "mycollection";
        final DefaultCollectionProfiler profiler = createProfiler(false);
        final DBCollection collection = mock(DBCollection.class);
        final String ns = "mydb." + collectionName;
        when(dbMock.collectionExists(collectionName)).thenReturn(true);
        when(dbMock.getCollection(collectionName)).thenReturn(collection);
        when(collection.getFullName()).thenReturn(ns);
        final CommandResult stats = mock(CommandResult.class);
        when(stats.get("count")).thenReturn(77L);
        when(stats.get("paddingFactor")).thenReturn(1.77);
        when(stats.get("size")).thenReturn(277L);
        when(stats.get("storageSize")).thenReturn(377L);
        when(collection.getStats()).thenReturn(stats);
        when(stats.ok()).thenReturn(true);
        profiler.addProfile(collectionName);

        // When
        profiler.newSample();

        // Then
        verify(collection).getStats();
        assertEquals(profiler.getProfiles().size(), 1);
        final DefaultCollectionProfile profile = profiler.getProfile(collectionName);
        assertNotNull(profile);
        assertEquals(profile.getCollectionName(), collectionName);

        // When
        profiler.reset();

        // Then
        assertEquals(profiler.getProfiles().size(), 1);
        assertNull(profiler.getProfile(ns));
    }


}
