package de.otto.mongodb.profiler.op;

import com.mongodb.BasicDBObject;
import com.mongodb.CommandResult;
import com.mongodb.DB;
import com.mongodb.DBObject;
import de.otto.mongodb.profiler.AbstractAsyncProfilerTest;
import de.otto.mongodb.profiler.op.analyze.Analyzer;
import org.mockito.ArgumentCaptor;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Collection;
import java.util.Collections;
import java.util.concurrent.ScheduledExecutorService;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.*;
import static org.testng.Assert.assertEquals;

public class DefaultOpProfilerTest extends AbstractAsyncProfilerTest {

    private DB dbMock;
    private ScheduledExecutorService fetcherServiceMock;
    private ScheduledExecutorService analyzerServiceMock;

    @BeforeMethod
    public void setUp() {
        dbMock = mock(DB.class);
        fetcherServiceMock = mock(ScheduledExecutorService.class);
        analyzerServiceMock = mock(ScheduledExecutorService.class);
    }

    @Override
    protected DefaultOpProfiler createProfiler(boolean execute) {
        final Collection<? extends Analyzer> analyzers = Collections.emptyList();
        return new DefaultOpProfiler(execute, dbMock, analyzers, fetcherServiceMock, analyzerServiceMock);
    }

    @Test
    public void shouldScheduleTasksWithDelayOnInitialization() {

        // Given
        final DefaultOpProfiler profiler = createProfiler(false);

        // When
        profiler.initialize();

        // Then
        verify(fetcherServiceMock).scheduleWithFixedDelay(
                any(OpProfileDataFetcher.class),
                eq(DefaultOpProfiler.DEFAULT_FETCHER_INITIAL_DELAY),
                eq(DefaultOpProfiler.DEFAULT_FETCHER_DELAY),
                eq(DefaultOpProfiler.DEFAULT_FETCHER_TIMEUNIT));
        verify(analyzerServiceMock).scheduleWithFixedDelay(
                any(OpProfileAnalyzer.class),
                eq(DefaultOpProfiler.DEFAULT_ANALYZER_INITIAL_DELAY),
                eq(DefaultOpProfiler.DEFAULT_ANALYZER_DELAY),
                eq(DefaultOpProfiler.DEFAULT_ANALYZER_TIMEUNIT));
    }

    @Test
    public void shouldShutdownExecutorServices() throws InterruptedException {

        // Given
        final DefaultOpProfiler profiler = createProfiler(false);
        immediateShutdown(fetcherServiceMock, analyzerServiceMock);

        // When
        profiler.shutdown();

        // Then
        verify(fetcherServiceMock).shutdown();
        verify(analyzerServiceMock).shutdown();
    }

    private static CommandResult profile(int level) {
        final CommandResult result = mock(CommandResult.class);
        when(result.getInt(eq("was"), anyInt())).thenReturn(level);
        return result;
    }

    @DataProvider
    public Object[][] shouldGetTheProfilingLevelData() {

        return new Object[][]{
                new Object[]{profile(0), OpProfiler.ProfilingLevel.OFF},
                new Object[]{profile(1), OpProfiler.ProfilingLevel.SLOW_OPS},
                new Object[]{profile(2), OpProfiler.ProfilingLevel.ALL_OPS}
        };
    }

    @Test(dataProvider = "shouldGetTheProfilingLevelData")
    public void shouldGetTheProfilingLevel(CommandResult dbo, OpProfiler.ProfilingLevel expectedLevel) {

        // Given
        when(dbMock.command(eq(new BasicDBObject("profile", -1)))).thenReturn(dbo);
        final OpProfiler profiler = createProfiler(false);

        // When
        final OpProfiler.ProfilingLevel result = profiler.getProfilingLevel();

        // Then
        assertEquals(result, expectedLevel);
    }

    @DataProvider
    public Object[][] shouldSetTheProfilingLevelData() {

        return new Object[][]{
                new Object[]{OpProfiler.ProfilingLevel.OFF, new BasicDBObject("profile", 0)},
                new Object[]{OpProfiler.ProfilingLevel.SLOW_OPS, new BasicDBObject("profile", 1)},
                new Object[]{OpProfiler.ProfilingLevel.ALL_OPS, new BasicDBObject("profile", 2)}
        };
    }

    @Test(dataProvider = "shouldSetTheProfilingLevelData")
    public void shouldSetTheProfilingLevel(OpProfiler.ProfilingLevel level, DBObject expectedCommand) {

        // Given
        final OpProfiler profiler = createProfiler(false);
        final CommandResult commandResult = mock(CommandResult.class);
        when(dbMock.command(any(DBObject.class))).thenReturn(commandResult);

        // When
        profiler.setProfilingLevel(level);

        // Then
        final ArgumentCaptor<DBObject> captor = ArgumentCaptor.forClass(DBObject.class);
        verify(dbMock).command(captor.capture());
        assertEquals(captor.getValue(), expectedCommand);
    }

}
