package de.otto.mongodb.profiler.op;

import com.mongodb.*;
import de.otto.mongodb.profiler.AbstractAsyncProfiler;
import de.otto.mongodb.profiler.FunctionTest;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Date;
import java.util.LinkedList;
import java.util.Queue;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Mockito.*;
import static org.testng.Assert.*;

public class OpProfileDataFetcherTest extends FunctionTest {

    private AbstractAsyncProfiler.ExecutionGuard executionGuardMock;
    private DB dbMock;
    private DBCollection systemProfileCollectionMock;
    private DBCursor systemProfileCursorMock;
    private Queue<DBObject> rawDataQueue;
    private OpProfileDataFetcher fetcher;

    private static class CursorAnswer implements Answer<Object> {

        private int invocationCount = 0;
        private final int maximumInvocations;

        private CursorAnswer(int maximumInvocations) {
            this.maximumInvocations = maximumInvocations;
        }

        @Override
        public Object answer(InvocationOnMock invocation) throws Throwable {
            return ++invocationCount <= maximumInvocations ? Boolean.TRUE : Boolean.FALSE;
        }
    }

    @BeforeMethod
    public void setUp() {

        executionGuardMock = mock(AbstractAsyncProfiler.ExecutionGuard.class);
        dbMock = mock(DB.class);
        systemProfileCollectionMock = mock(DBCollection.class);
        systemProfileCursorMock = mock(DBCursor.class);
        rawDataQueue = new LinkedList<>();
        fetcher = new OpProfileDataFetcher(executionGuardMock, dbMock, rawDataQueue);

        when(dbMock.getCollection(OpProfileDataFetcher.COLLECTION)).thenReturn(systemProfileCollectionMock);
        when(dbMock.collectionExists(OpProfileDataFetcher.COLLECTION)).thenReturn(true);
        when(systemProfileCollectionMock.find(any(DBObject.class))).thenReturn(systemProfileCursorMock);
        when(systemProfileCursorMock.addOption(anyInt())).thenReturn(systemProfileCursorMock);
        when(systemProfileCursorMock.sort(any(DBObject.class))).thenReturn(systemProfileCursorMock);
        when(systemProfileCursorMock.batchSize(anyInt())).thenReturn(systemProfileCursorMock);
    }

    @Test
    public void shouldNotExecuteWhenExecutionGuardDeniesIt() {

        // Given
        when(executionGuardMock.execute()).thenReturn(false);
        when(dbMock.collectionExists(OpProfileDataFetcher.COLLECTION)).thenReturn(true);

        // When
        fetcher.run();

        // Then
        verifyZeroInteractions(systemProfileCollectionMock, systemProfileCursorMock);
    }

    @Test
    public void shouldNotExecuteWhenProfilingCollectionDoesNotExist() {

        // Given
        when(executionGuardMock.execute()).thenReturn(true);
        when(dbMock.collectionExists(OpProfileDataFetcher.COLLECTION)).thenReturn(false);

        // When
        fetcher.run();

        // Then
        verifyZeroInteractions(systemProfileCollectionMock, systemProfileCursorMock);
    }

    @Test
    public void shouldPutNothingOnQueueWhenCursorHasntSomething() {

        // Given
        when(executionGuardMock.execute()).thenReturn(true);
        when(systemProfileCursorMock.hasNext()).thenReturn(false);

        // When
        fetcher.run();

        // Then
        verify(dbMock).getCollection(OpProfileDataFetcher.COLLECTION);
        verify(systemProfileCollectionMock).find(any(DBObject.class));
        assertTrue(rawDataQueue.isEmpty());
    }

    @Test
    public void shouldPutDataOnQueue() {

        // Given
        when(executionGuardMock.execute()).thenReturn(true);
        when(systemProfileCursorMock.hasNext()).thenAnswer(new CursorAnswer(1));
        final DBObject data = BasicDBObjectBuilder
                .start()
                .add("ns", "someNs")
                .add("ts", new Date())
                .get();
        when(systemProfileCursorMock.next()).thenReturn(data);

        // When
        fetcher.run();

        // Then
        verify(dbMock).getCollection(OpProfileDataFetcher.COLLECTION);
        verify(systemProfileCollectionMock).find(any(DBObject.class));
        assertFalse(rawDataQueue.isEmpty());
        assertSame(rawDataQueue.remove(), data);
        assertTrue(rawDataQueue.isEmpty());
    }

    @DataProvider
    public Object[][] shouldIgnoreNamespacesData() {

        return new Object[][]{
                new Object[]{"somedb.system.namespaces"},
                new Object[]{"somedb.system.indexes"},
                new Object[]{"somedb.system.profile"},
                new Object[]{"somedb.system.js"}
        };
    }

    @Test(dataProvider = "shouldIgnoreNamespacesData")
    public void shouldIgnoreNamespaces(String namespace) {

        // Given
        when(executionGuardMock.execute()).thenReturn(true);
        when(systemProfileCursorMock.hasNext()).thenAnswer(new CursorAnswer(1));
        final DBObject data = BasicDBObjectBuilder
                .start()
                .add("ns", namespace)
                .add("ts", new Date())
                .get();
        when(systemProfileCursorMock.next()).thenReturn(data);

        // When
        fetcher.run();

        // Then
        verify(dbMock).getCollection(OpProfileDataFetcher.COLLECTION);
        verify(systemProfileCollectionMock).find(any(DBObject.class));
        assertTrue(rawDataQueue.isEmpty());
    }

    @Test
    public void shouldCloseACursor() {

        // Given
        when(executionGuardMock.execute()).thenReturn(true);
        when(systemProfileCursorMock.hasNext()).thenReturn(false);
        fetcher.run();
        verify(systemProfileCursorMock, never()).close();

        // When
        fetcher.close();

        // Then
        verify(systemProfileCursorMock).close();
    }

}
