package de.otto.mongodb.profiler.util;

import com.mongodb.DBObject;
import de.otto.mongodb.profiler.FunctionTest;
import org.testng.annotations.Test;

import static de.otto.mongodb.profiler.util.MongoSupport.get;
import static de.otto.mongodb.profiler.util.MongoSupport.require;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertSame;

public class MongoSupportTest extends FunctionTest {

    @Test
    public void shouldNotReturnValueThatIsNotFound() {

        // Given
        final DBObject dbo = mock(DBObject.class);

        // When
        final String result = get("field", String.class, dbo);

        // Then
        assertNull(result);
    }

    @Test
    public void shouldReturnValueThatIsFound() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final String foo = "foo";
        when(dbo.get("field")).thenReturn(foo);

        // When
        final String result = get("field", String.class, dbo);

        // Then
        assertSame(result, foo);
    }

    @Test
    public void shouldReturnDefaultValueWhenValueIsNotFound() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final String defaultValue = "foo";

        // When
        final String result = get("field", String.class, dbo, defaultValue);

        // Then
        assertSame(result, defaultValue);
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void shouldFailIfTypeIsDifferent() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final Integer foo = Integer.valueOf(1);
        when(dbo.get("field")).thenReturn(foo);

        // When
        get("field", String.class, dbo);

        // Then
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void shouldFailIfTypeIsDifferentWithDefaultValue() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final Integer foo = Integer.valueOf(1);
        when(dbo.get("field")).thenReturn(foo);
        final String defaultValue = "foo";

        // When
        get("field", String.class, dbo, defaultValue);

        // Then
    }

    @Test
    public void shouldReturnRequiredValueThatIsFound() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final String foo = "foo";
        when(dbo.get("field")).thenReturn(foo);

        // When
        final String result = require("field", String.class, dbo);

        // Then
        assertSame(result, foo);
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void shouldFailWhenRequiredValueIsNotFound() {

        // Given
        final DBObject dbo = mock(DBObject.class);

        // When
        require("field", String.class, dbo);

        // Then
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void shouldFailWhenRequiredValueIsOfDifferentType() {

        // Given
        final DBObject dbo = mock(DBObject.class);
        final Integer foo = Integer.valueOf(1);
        when(dbo.get("field")).thenReturn(foo);

        // When
        require("field", String.class, dbo);

        // Then
    }
}
